from functools import wraps
from typing import TYPE_CHECKING, Optional

from rich.console import Console
from rich.markup import escape
from sqlmodel import Session

from synda.model.node import Node, NodeStatus
from synda.database import engine
from synda.model.run import Run, RunStatus
from synda.model.step import Step
from synda.utils.env import is_debug_enabled

if TYPE_CHECKING:
    from synda.config import Config

CONSOLE = Console()


class Pipeline:
    def __init__(self, config: Optional["Config"] = None):
        self.session = Session(engine)
        self.config = config
        self.input_loader = config.input.get_loader() if config else None
        self.output_saver = config.output.get_saver() if config else None
        self.pipeline = config.pipeline if config else None
        self.run = Run.create_with_steps(self.session, config) if config else None

    @staticmethod
    def handle_run_errors(func):
        """Decorator to handle run errors gracefully."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except Exception as e:
                if self.run is not None:
                    self.run.update(self.session, RunStatus.ERRORED)
                raise e

        return wrapper

    @staticmethod
    def handle_stop_option(func):
        """Decorator to handle user interruption (Ctrl+C) during execution."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except KeyboardInterrupt:
                if self.run is not None:
                    prompt = f"\nAre you sure you want to stop the run {self.run.id}? [y/N]: "
                    escaped_prompt = escape(prompt)

                    user_input = (
                        CONSOLE.input(f"[red]{escaped_prompt}[/]").strip().lower()
                    )

                    if user_input == "y":
                        self.run.update(self.session, RunStatus.STOPPED)
                        CONSOLE.print(
                            f"[blue]Run with id {self.run.id} is stopped.\nTo resume the run, use:"
                        )
                        CONSOLE.print(f"[cyan]synda generate --resume {self.run.id}")
                        exit(0)
                    else:
                        self.resume(run_id=self.run.id)

        return wrapper

    @handle_run_errors
    @handle_stop_option
    def execute(self):
        if self.config is None:
            raise ValueError("Config can't be None to execute a pipeline")

        input_nodes = self.input_loader.load(self.session)

        for step in self.run.steps:
            self._log_debug_info(step)
            executor = step.get_step_config().get_executor(self.session, self.run, step)
            input_nodes = executor.execute_and_update_step(input_nodes, [], False)

        self._finalize_run(input_nodes)

    @handle_run_errors
    @handle_stop_option
    def retry(self):
        CONSOLE.print("[blue]Retrying last failed run")
        last_failed_step = Step.get_last_failed(self.session)

        if last_failed_step is None:
            raise Exception("Can't find any failed step.")

        self._restart_from_step(last_failed_step)

    @handle_run_errors
    @handle_stop_option
    def resume(self, run_id: int):
        CONSOLE.print(f"[blue]Resuming run {run_id}")
        resumed_step = Step.get_step_to_resume(session=self.session, run_id=run_id)
        self._restart_from_step(resumed_step)

    def _restart_from_step(self, step: Step):
        from synda.config import Config

        self.run, input_nodes, remaining_steps = Run.restart_from_step(
            session=self.session, step=step
        )
        self.config = Config.model_validate(self.run.config)
        self.output_saver = self.config.output.get_saver()

        self._execute_remaining_steps(input_nodes, remaining_steps)

        self._finalize_run(input_nodes)

    def _execute_remaining_steps(self, input_nodes, remaining_steps):
        is_first_remaining_step = True

        for current_step in remaining_steps:
            self._log_debug_info(current_step)
            executor = current_step.get_step_config().get_executor(
                self.session, self.run, current_step
            )

            nodes_to_process, processed_nodes = self._prepare_nodes(
                input_nodes, is_first_remaining_step
            )

            input_nodes = executor.execute_and_update_step(
                nodes_to_process, processed_nodes, is_first_remaining_step
            )
            is_first_remaining_step = False

        return input_nodes

    @staticmethod
    def _prepare_nodes(input_nodes, is_first_step):
        if not is_first_step:
            return input_nodes, []

        pending_nodes, processed_nodes = [], []

        for node in input_nodes:
            if node.status == NodeStatus.PENDING:
                pending_nodes.append(node)
            elif node.status == NodeStatus.PROCESSED:
                processed_nodes.append(node)

        return pending_nodes, processed_nodes

    @staticmethod
    def _log_debug_info(step):
        if is_debug_enabled():
            print(step)

    def _finalize_run(self, input_nodes: list[Node]):
        self.output_saver.save(input_nodes)
        self.run.update(self.session, RunStatus.FINISHED)
        CONSOLE.print(f"[green]Run {self.run.id} finished successfully!")



import re
import unicodedata

from sqlmodel import Session

from synda.model.run import Run
from synda.pipeline.executor import Executor
from synda.model.node import Node
from synda.progress_manager import ProgressManager
from synda.model.step import Step
from synda.utils.prompt_builder import PromptBuilder


class WordPosition(Executor):
    def __init__(self, session: Session, run: Run, step_model: Step):
        super().__init__(session, run, step_model)
        self.progress = ProgressManager("METADATA")

    def execute(self, pending_nodes: list[Node], processed_nodes: list[Node]):
        result = []
        matches = self.config.parameters.matches

        with self.progress.task("  Metadata...", len(pending_nodes)) as advance:
            for node in pending_nodes:
                metadata = []
                text = node.value

                for label, pattern in matches.items():
                    pattern = PromptBuilder.build(self.session, pattern, [node])
                    regex_pattern = self._create_pattern_ignoring_case_and_accents(
                        pattern
                    )
                    match = re.search(regex_pattern, text)

                    if match:
                        metadata_entry = {
                            "label": label,
                            "start": match.start(),
                            "end": match.end(),
                            "value": match.group(),
                        }
                        metadata.append(metadata_entry)

                result.append(
                    Node(
                        parent_node_id=node.id, value=node.value, node_metadata=metadata
                    )
                )
                advance()

        return result

    @staticmethod
    def _create_pattern_ignoring_case_and_accents(pattern: str) -> str:
        pattern = pattern.lower()
        pattern = unicodedata.normalize("NFKD", pattern)
        pattern = "".join(c for c in pattern if ord(c) < 128)
        pattern = re.escape(pattern)
        return f"(?i){pattern}"



from abc import abstractmethod

from sqlmodel import Session

from synda.model.run import Run
from synda.model.step import Step, StepStatus
from synda.model.node import Node


class Executor:
    def __init__(
        self,
        session: Session,
        run: Run,
        step_model: Step,
        save_on_completion: bool = True,
    ):
        self.session = session
        self.run = run
        self.step_model = step_model
        self.config = step_model.get_step_config()
        self.save_on_completion = save_on_completion

    def execute_and_update_step(
        self,
        pending_nodes: list[Node],
        processed_nodes: list[Node],
        restarted_step: bool = False,
    ) -> list[Node]:
        try:
            self.step_model.set_running(
                self.session, pending_nodes, restarted=restarted_step
            )

            output_nodes = self.execute(pending_nodes, processed_nodes)
            if self.save_on_completion:
                self.step_model.save_at_execution_end(
                    self.session, pending_nodes, output_nodes
                )
            else:
                self.step_model.set_completed(session=self.session)

            filtered_nodes = [node for node in output_nodes if not node.ablated]

            return filtered_nodes
        except Exception as e:
            self.step_model.set_status(self.session, StepStatus.ERRORED)
            raise e

    @abstractmethod
    def execute(self, pending_nodes: list[Node], processed_nodes: list[Node]):
        pass



import random
import re
from enum import Enum

from sqlmodel import Session

from synda.model.node import Node


class SpecialVariable(Enum):
    INSTRUCTIONS = "instructions"


class PromptBuilder:
    # @todo send multiple template at once: template: str | list[str
    # @todo send a single node at once: input_data: list[Node] | Node
    @staticmethod
    def build(
        session: Session,
        template: str,
        input_data: list[Node],
        instruction_sets: dict[str, list[str]] | None = None,
    ) -> list[str] | str:
        prompts_with_special_variables = (
            PromptBuilder._build_prompts_with_special_variables(
                template, input_data, instruction_sets
            )
        )
        prompts_with_template_variables = (
            PromptBuilder._build_prompts_with_template_variables(
                session, prompts_with_special_variables, input_data
            )
        )

        return (
            prompts_with_template_variables
            if len(prompts_with_template_variables) > 1
            else prompts_with_template_variables[0]
        )

    @staticmethod
    def _build_prompts_with_special_variables(
        template: str,
        input_data: list[Node],
        instruction_sets: dict[str, list[str]] | None,
    ) -> list[str]:
        special_variables = PromptBuilder._extract_special_variables(template)

        if len(special_variables) == 0:
            return [template for _ in input_data]

        prompts_with_special_variables = []

        for _ in input_data:
            variable_value = {}

            for variable_name in special_variables:
                if variable_name == SpecialVariable.INSTRUCTIONS.value:
                    variable_value[SpecialVariable.INSTRUCTIONS.value] = (
                        PromptBuilder._build_instructions(instruction_sets)
                    )

            formatted_prompt = template.format(**variable_value)
            prompts_with_special_variables.append(formatted_prompt)

        return prompts_with_special_variables

    @staticmethod
    def _build_prompts_with_template_variables(
        session: Session,
        prompts_with_special_variables: list[str],
        input_data: list[Node],
    ) -> list[str]:
        template_variables = set()
        for prompt in prompts_with_special_variables:
            template_variables.update(PromptBuilder._extract_template_variables(prompt))

        if len(template_variables) == 0:
            return prompts_with_special_variables

        prompts_with_template_variables = []

        parent_nodes = PromptBuilder._get_parent_nodes(
            session, template_variables, input_data
        )

        for node, prompt in zip(input_data, prompts_with_special_variables):
            variable_value = {}

            for variable_name in template_variables:
                parent_node_id = node.ancestors[variable_name]
                parent_node = next(
                    node for node in parent_nodes if node.id == parent_node_id
                )
                variable_value[variable_name] = parent_node.value

            formatted_prompt = prompt.format(**variable_value)
            prompts_with_template_variables.append(formatted_prompt)

        return prompts_with_template_variables

    @staticmethod
    def _extract_special_variables(template: str) -> list[str]:
        pattern = PromptBuilder._pattern_only_special_variables()
        matches = re.finditer(pattern, template)
        return [match.group(1) for match in matches]

    @staticmethod
    def _extract_template_variables(template: str) -> list[str]:
        pattern = PromptBuilder._pattern_without_special_variables()
        matches = re.finditer(pattern, template)
        return [match.group(1) for match in matches]

    @staticmethod
    def _pattern_only_special_variables() -> str:
        special_vars = "|".join(var.value for var in SpecialVariable)
        return rf"{{({special_vars})}}"

    @staticmethod
    def _pattern_without_special_variables() -> str:
        special_vars = "|".join(var.value for var in SpecialVariable)
        return rf"{{(?!{special_vars})([^}}]+)}}"

    @staticmethod
    def _build_instructions(instruction_sets: dict[str, list[str]] | None) -> str:
        if instruction_sets is None:
            raise Exception("`instruction_sets` is missing from parameters.")

        instructions_set = []

        for category, instructions in instruction_sets.items():
            selected_instruction = random.choice(instructions)
            instructions_set.append("- " + selected_instruction)

        return "\n".join(instructions_set)

    @staticmethod
    def _get_parent_nodes(
        session: Session, variables: set[str], input_data: list[Node]
    ) -> list[Node]:
        parent_node_ids = set()

        for node in input_data:
            for variable_name in variables:
                parent_node_ids.add(node.ancestors[variable_name])

        return Node.get(session, list(parent_node_ids))



from pathlib import Path
import typer
from typing import Optional

from synda.config import Config
from synda.pipeline import Pipeline


def generate_command(
    config_file: Path = typer.Argument(
        default=None,
        help="Path to YAML configuration file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    retry: bool = typer.Option(
        False, "--retry", "-r", help="Run the pipeline from last failed step"
    ),
    run_id: Optional[int] = typer.Option(
        None, "--resume", "-re", help="Resume the pipeline from a given run id"
    ),
):
    """Run a pipeline with provided configuration."""
    if retry:
        Pipeline().retry()
    elif run_id is not None:
        Pipeline().resume(run_id=run_id)
    else:
        config = Config.load_config(config_file)
        pipeline = Pipeline(config)
        pipeline.execute()



from typing import Annotated, Literal

from pydantic import BaseModel, Field, TypeAdapter
from sqlmodel import Session

from synda.config.step import Step
from synda.model.run import Run
from synda.model.step import Step as StepModel
from synda.pipeline.executor import Executor


class WordPositionParameters(BaseModel):
    matches: dict[str, str] = Field(
        description="A dictionary with 'key' = LABEl and 'value' = string to match"
    )


class WordPosition(Step):
    type: str = "metadata"
    method: Literal["word-position"]
    parameters: WordPositionParameters

    def get_executor(
        self, session: Session, run: Run, step_model: StepModel
    ) -> Executor:
        from synda.pipeline.metadata import WordPosition

        return WordPosition(session, run, step_model)


Metadata = Annotated[WordPosition, Field(discriminator="method")]

metadata_adapter = TypeAdapter(Metadata)



from enum import Enum
from typing import TYPE_CHECKING, Union, Any
from sqlmodel import SQLModel, Field, Relationship, Column, JSON, Session, select
from sqlalchemy import and_

from synda.model.step_node import StepNode, StepNodeRelationshipType

if TYPE_CHECKING:
    from synda.model.step import Step


class NodeStatus(Enum):
    PENDING = "pending"
    PROCESSED = "processed"


class Node(SQLModel, table=True):
    id: int | None = Field(default=None, primary_key=True)
    parent_node_id: int | None = None
    ablated: bool = False
    value: str
    ancestors: dict = Field(default_factory=dict, sa_column=Column(JSON))
    status: NodeStatus = Field(default=NodeStatus.PENDING)
    node_metadata: list[dict[str, Any]] = Field(
        default_factory=list, sa_column=Column(JSON)
    )

    step_node_links: list["StepNode"] = Relationship(back_populates="node")

    input_step: list["Step"] = Relationship(
        back_populates="input_nodes",
        link_model=StepNode,
        sa_relationship_kwargs={
            "primaryjoin": f"and_(Node.id == StepNode.node_id, StepNode.relationship_type == '{StepNodeRelationshipType.INPUT.value}')",  # noqa
            "secondaryjoin": "Step.id == StepNode.step_id",
            "secondary": "step_node",
            "overlaps": "step,node,step_node_links",
        },
    )

    output_step: list["Step"] = Relationship(
        back_populates="output_nodes",
        link_model=StepNode,
        sa_relationship_kwargs={
            "primaryjoin": f"and_(Node.id == StepNode.node_id, StepNode.relationship_type == '{StepNodeRelationshipType.OUTPUT.value}')",  # noqa
            "secondaryjoin": "Step.id == StepNode.step_id",
            "secondary": "step_node",
            "overlaps": "input_step,node,step_node_links,step",
        },
    )

    @classmethod
    def get(
        cls, session: Session, node_ids: Union[int, list[int]]
    ) -> Union["Node", list["Node"]]:
        if isinstance(node_ids, int):
            node_ids = [node_ids]
            single_result = True
        else:
            single_result = False

        query = select(cls).where(cls.id.in_(node_ids))
        results = session.exec(query).all()

        return results[0] if single_result and results else results

    @staticmethod
    def get_input_nodes_from_step(session: Session, step: "Step") -> list["Node"]:
        return session.exec(
            select(Node)
            .join(StepNode, Node.id == StepNode.node_id)
            .where(
                and_(
                    StepNode.step_id == step.id,
                    StepNode.relationship_type == StepNodeRelationshipType.INPUT.value,
                )
            )
        ).fetchall()

    def is_ablated_text(self) -> str:
        return "yes" if self.ablated else "no"

    def set_processed(self, session: Session):
        self.status = NodeStatus.PROCESSED
        session.add(self)
        session.commit()
        session.refresh(self)



from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import TfidfVectorizer
from sqlmodel import Session

from synda.model.run import Run
from synda.pipeline.executor import Executor
from synda.model.node import Node
from synda.progress_manager import ProgressManager
from synda.model.step import Step


class DeduplicateTFIDF(Executor):
    def __init__(self, session: Session, run: Run, step_model: Step):
        super().__init__(session, run, step_model)
        self.progress = ProgressManager("CLEAN")

    def execute(
        self, pending_nodes: list[Node], processed_nodes: list[Node]
    ) -> list[Node]:
        strategy = self.config.parameters.strategy
        similarity_threshold = self.config.parameters.similarity_threshold
        keep = self.config.parameters.keep

        with self.progress.task(
            "  Cleaning...",
            len(pending_nodes) + len(processed_nodes),
            completed=len(processed_nodes),
        ) as advance:
            if strategy == "exact":
                result_nodes = self._remove_exact_duplicates(
                    pending_nodes, keep, advance
                )
            elif strategy == "fuzzy":
                result_nodes = self._remove_fuzzy_duplicates(
                    pending_nodes, similarity_threshold, keep, advance
                )

        return [Node(parent_node_id=node.id, value=node.value) for node in result_nodes]

    @staticmethod
    def _remove_exact_duplicates(
        pending_nodes: list[Node], keep: str, advance
    ) -> list[Node]:
        seen_values = set()
        result = []

        if keep == "first":
            for node in pending_nodes:
                if node.value not in seen_values:
                    seen_values.add(node.value)
                    result.append(node)
                advance()
        elif keep == "last":
            for node in reversed(pending_nodes):
                if node.value not in seen_values:
                    seen_values.add(node.value)
                    result.insert(0, node)
                advance()

        return result

    @staticmethod
    def _remove_fuzzy_duplicates(
        pending_nodes: list[Node], similarity_threshold: float, keep: str, advance
    ) -> list[Node]:
        node_values = [node.value for node in pending_nodes]

        vectorizer = TfidfVectorizer(strip_accents="unicode")
        tfidf_matrix = vectorizer.fit_transform(node_values)
        similarity_matrix = cosine_similarity(tfidf_matrix)
        index_node_to_keep = set(range(len(node_values)))

        for node_index in range(len(node_values)):
            if node_index not in index_node_to_keep:
                advance()
                continue

            for compared_node_index in range(node_index + 1, len(node_values)):
                if compared_node_index not in index_node_to_keep:
                    continue
                if (
                    similarity_matrix[node_index, compared_node_index]
                    > similarity_threshold
                ):
                    if keep == "first":
                        index_node_to_keep.discard(compared_node_index)
                    elif keep == "last":
                        index_node_to_keep.discard(node_index)
                        break

            advance()

        return [pending_nodes[i] for i in sorted(index_node_to_keep)]



import json
from typing import Literal, Optional

from pydantic import BaseModel
from sqlmodel import Session

from synda.model.provider import Provider
from synda.model.run import Run
from synda.model.step import Step
from synda.pipeline.executor import Executor
from synda.model.node import Node
from synda.progress_manager import ProgressManager
from synda.utils.env import is_debug_enabled
from synda.utils.llm_provider import LLMProvider
from synda.utils.prompt_builder import PromptBuilder


class LLMJudgeCriterionBinaryAnswer(BaseModel):
    # @todo hotfix structured generation with ollama doesn't work as expected
    answer: Literal["YES", "NO", "OUI", "NON"]

    def is_positive_answer(self) -> bool:
        return self.answer == "YES" or self.answer == "OUI"


class LLMJudgeBinary(Executor):
    def __init__(self, session: Session, run: Run, step_model: Step):
        super().__init__(session, run, step_model, save_on_completion=False)
        self.progress = ProgressManager("ABLATION")
        self.provider = Provider.get(self.config.parameters.provider)
        self.model = self.config.parameters.model

    # @todo build criteria prompts in a single call
    def execute(self, pending_nodes: list[Node], processed_nodes: list[Node]):
        criteria = self.config.parameters.criteria
        result = processed_nodes or []

        with self.progress.task(
            "  Ablating...",
            (len(pending_nodes) + len(processed_nodes)) * len(criteria),
            completed=len(processed_nodes) * len(criteria),
        ) as advance_node:
            for node in pending_nodes:
                judge_answers = []
                for criterion in criteria:
                    criterion_prompt = PromptBuilder.build(
                        self.session, criterion, [node]
                    )
                    prompt = self._build_binary_judge_prompt(
                        node.value, criterion_prompt
                    )
                    judge_answer = LLMProvider.call(
                        self.provider.name,
                        self.model,
                        self.provider.api_key,
                        prompt,
                        LLMJudgeCriterionBinaryAnswer,
                        url=self.provider.api_url,
                        temperature=self.config.parameters.temperature,
                        format="json",
                    )
                    try:
                        judge_answer = LLMJudgeCriterionBinaryAnswer(
                            **json.loads(judge_answer)
                        )
                    except Exception as e:
                        print(f"Model response format is malformed: {e}")

                    judge_answers.append(judge_answer)
                    advance_node()

                ablated = not self._check_consensus(judge_answers)
                result_node = Node(
                    parent_node_id=node.id, value=node.value, ablated=ablated
                )
                self.step_model.save_during_execution(self.session, node, result_node)
                result.append(result_node)

            if is_debug_enabled():
                print(f"Answer: {node.value}")
                print(f"Criteria: {str(criteria)}")
                print(f"Consensus: {self.config.parameters.consensus}")
                print(f"Judge answers: {judge_answers}")
                print(f"Ablated: {result_node.is_ablated_text()}\n")

        return result

    def _check_consensus(
        self, judge_answers: list[LLMJudgeCriterionBinaryAnswer]
    ) -> bool:
        consensus = self.config.parameters.consensus.lower()
        positive_answers = sum(
            1 for answer in judge_answers if answer.is_positive_answer()
        )
        total_answers = len(judge_answers)

        if total_answers == 0:
            return False

        match consensus:
            case "all":
                return positive_answers == total_answers
            case "any":
                return positive_answers > 0
            case "majority":
                return positive_answers > total_answers / 2
            case _:
                raise ValueError(f"Unknown consensus: {consensus}")

    # @todo move the criterion into step parameters
    @staticmethod
    def _build_binary_judge_prompt(candidate: str, criterion: str) -> str:
        return (
            f"You are an expert judge tasked with evaluating synthetic text data.\n"
            f"You are evaluating synthetic data against a given criterion.\n"
            'You must answer ONLY with {"answer": "YES"} or {"answer": "NO"}.\n'
            'ALWAYS answer in english {"answer": "YES"} or {"answer": "NO"}, never in another language.\n'
            'Output {"answer": "YES"} when the criterion is fulfilled.\n'
            'Output {"answer": "NO"} when the criterion is NOT fulfilled.\n'
            f"------\n"
            f"criterion: Is the candidate written in english?\n"
            f"candidate: Great Britain is a bit pretentious to call itself “Great”.\n"
            '{"answer": "YES"}\n'
            f"------\n"
            f"criterion: Does the text contain more than 10 words?\n"
            f"candidate: The cat sleeps.\n"
            '{"answer": "NO"}\n'
            f"------\n"
            f"criterion: Does the text talks about the sun?\n"
            f"candidate: Synda is a synthetic data library.\n"
            '{"answer": "NO"}\n'
            f"------\n"
            f"criterion: {criterion}\n"
            f"candidate: {candidate}\n"
        )



from sqlmodel import Session

from synda.model.provider import Provider
from synda.model.run import Run
from synda.model.step import Step
from synda.pipeline.executor import Executor
from synda.model.node import Node
from synda.utils.llm_provider import LLMProvider
from synda.utils.prompt_builder import PromptBuilder
from synda.progress_manager import ProgressManager


class LLM(Executor):
    def __init__(self, session: Session, run: Run, step_model: Step):
        super().__init__(session, run, step_model, save_on_completion=False)
        self.progress = ProgressManager("GENERATION")
        self.provider = Provider.get(self.config.parameters.provider)
        self.model = self.config.parameters.model

    def execute(self, pending_nodes: list[Node], processed_nodes: list[Node]):
        template = self.config.parameters.template
        occurrences = self.config.parameters.occurrences
        instruction_sets = self.config.parameters.instruction_sets
        pending_nodes = self._build_node_occurrences(pending_nodes, occurrences)
        prompts = PromptBuilder.build(
            self.session, template, pending_nodes, instruction_sets=instruction_sets
        )
        result = processed_nodes
        with self.progress.task(
            "Generating...",
            len(pending_nodes) + len(processed_nodes),
            completed=len(processed_nodes),
        ) as advance:
            for node, prompt in zip(pending_nodes, prompts):
                llm_answer = LLMProvider.call(
                    self.provider.name,
                    self.model,
                    self.provider.api_key,
                    prompt,
                    url=self.provider.api_url,
                    temperature=self.config.parameters.temperature,
                )
                result_node = Node(parent_node_id=node.id, value=llm_answer)
                self.step_model.save_during_execution(self.session, node, result_node)
                result.append(result_node)
                advance()

        return result

    @staticmethod
    def _build_node_occurrences(
        pending_nodes: list[Node], occurrences: int
    ) -> list[Node]:
        nodes = []

        for node in pending_nodes:
            [nodes.append(node) for _ in range(occurrences)]

        return nodes



from enum import Enum
from typing import Literal, TYPE_CHECKING

from sqlmodel import SQLModel, Field, Relationship

if TYPE_CHECKING:
    from synda.config.step import Step
    from synda.model.node import Node


class StepNodeRelationshipType(Enum):
    INPUT = "input"
    OUTPUT = "output"


class StepNode(SQLModel, table=True):
    __tablename__ = "step_node"

    step_id: int = Field(foreign_key="step.id", primary_key=True)
    node_id: int = Field(foreign_key="node.id", primary_key=True)
    relationship_type: StepNodeRelationshipType = Field(index=True)

    step: "Step" = Relationship(back_populates="step_node_links")
    node: "Node" = Relationship(back_populates="step_node_links")



from typing import Any

import pandas as pd

from synda.config.output import Output
from synda.model.node import Node
from synda.pipeline.output.output_saver import OutputSaver


class CSVOutputSaver(OutputSaver):
    def __init__(self, output_config: Output):
        self.properties = output_config.properties
        super().__init__()

    def save(self, input_data: list[Node]) -> None:
        data: dict[str, list] = {}

        for column in self.properties.columns:
            data[column] = [
                self._get_node_attribute(node, column) for node in input_data
            ]

        df = pd.DataFrame(data)

        df.to_csv(self.properties.path, sep=self.properties.separator, index=False)

    @staticmethod
    def _get_node_attribute(node: Node, column: str) -> Any:
        if column == "value":
            return node.value
        elif column == "ablated":
            return node.is_ablated_text()
        elif column == "metadata":
            return node.node_metadata
        else:
            raise ValueError(f"Unknown column: {column}")
